import os
import sys
import json
import importlib.util
import time
import threading
import queue
import datetime
import multiprocessing
import traceback
from pg_tasks_queue.Config import cfg as config
import pg_tasks_queue.Database as database
from pg_tasks_queue.Kthread import KThread as kthread

import logging
logger = logging.getLogger(__name__)


class Worker:

    _blocking = True
    _sleep_sec = None
    _timeout_sec = 600.
    _life_timeout_sec = None
    _started = False
    _started_timestamp = None
    _worker = None
    _current_task = None
    _raise_exception = False

    def check_module(self, module_name):
        func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
        module_spec = importlib.util.find_spec(module_name)
        if module_spec is None:
            logger.error(f'Error in {func_name}: module: "{module_name}" not found')
            return None
        else:
            return module_spec

    def import_module(self, module_name):
        func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
        module_spec = self.check_module(module_name)
        if module_spec is None:
            return None
        module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)
        return module

    def import_function_from_module(self, module_name, function_name):
        func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
        module = self.import_module(module_name)
        if module is None:
            return None
        if not hasattr(module, function_name):
            logger.error(f'Error in {func_name}: module: "{module_name}"; function: "{function_name}" not found')
            return None
        return getattr(module, function_name)

    def worker(self, task_dict):
        func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
        # print(f'{func_name}; os.getpid(): {os.getpid()}; os.getppid(): {os.getppid()}')
        task_module = task_dict.get('module')
        task_func = task_dict.get('func')
        func = self.import_function_from_module(task_module, task_func)
        if func is None:
            error = f'Error in {func_name}: import_function_from_module({task_module}, {task_func}) is None'
            logger.error(error)
            return {'status': 'error', 'result': error}
        else:
            params = task_dict.get('params')
            if isinstance(params, str):
                params = json.loads(params)
            if not isinstance(params, dict):
                params = dict()
            logger.info(f'Started task (id={task_dict.get("task_id")}) at {datetime.datetime.now()}; task_dict: {task_dict}')
            res = func(**params)
            logger.info(f'Finished task (id={task_dict.get("task_id")}) at {datetime.datetime.now()}; result: {res}')
            return res

    def worker_process(self, queue, task_dict):
        result = self.worker(task_dict)
        queue.put(result)

    def stop_woker(self, func_name):
        error = f'Error in {func_name}: worker.is_alive() => os.kill(os.getpid(), signal.SIGINT)'
        if self._worker:
            database.set_worker_error(error, self._worker, self._current_task)
        try:
            # sys.exit(1)
            import signal
            os.kill(os.getpid(), signal.SIGINT)
        finally:
            logger.error(error)

    def _do_task(self, process_type):
        func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
        # print(f'{func_name}; os.getpid(): {os.getpid()}; os.getppid(): {os.getppid()}')
        if not self._started:
            logger.warning(f'{func_name}; self._started = False; return')
            return

        exception = None
        error = None
        timer = None
        try:
            self._current_task = database.get_new_task(self._worker)
            if isinstance(self._current_task, str):
                logger.error(f'Error in {func_name}; {self._current_task}')
                return
            elif self._current_task is None:
                return

            # print(f'{func_name}: now={datetime.datetime.now()}; task: {self._current_task}')
            task_dict = database.get_record_dict(self._current_task)
            if process_type is None:
                # timer = threading.Timer(self._timeout_sec, self.stop_woker, args=(func_name,))
                # timer.setDaemon(False)
                # timer.start()
                result_dict = self.worker(task_dict)
                # timer.cancel()
                database.update_worker_task(result_dict, self._current_task, self._worker)
            elif process_type in ['fork', 'thread', 'kthread']:
                if process_type == 'fork':
                    worker_queue = multiprocessing.Queue()
                    worker = multiprocessing.Process(target=self.worker_process, args=(worker_queue, task_dict,))
                else:
                    worker_queue = queue.Queue()
                    if process_type == 'thread':
                        worker = threading.Thread(target=self.worker_process, args=(worker_queue, task_dict,))
                        worker.setDaemon(True)
                    else:
                        worker = kthread(target=self.worker_process, args=(worker_queue, task_dict,))
                worker.start()
                worker.join(timeout=self._timeout_sec)
                if worker.is_alive():
                    if process_type in ['fork', 'kthread']:
                        worker.terminate()
                        error = f'Error in {func_name}: worker({process_type}).is_alive() => worker.terminate()'
                        logger.error(error)
                        database.update_worker_task({'status': 'error', 'result': error}, self._current_task,
                                                    self._worker)
                    else:
                        error = f'Error in {func_name}: worker({process_type}).is_alive() => sys.exit(1)'
                        logger.error(error)
                        if self._worker:
                            database.set_worker_error(error, self._worker, self._current_task)
                        sys.exit(1)
                else:
                    if not worker_queue.empty():
                        database.update_worker_task(worker_queue.get_nowait(), self._current_task, self._worker)
                    else:
                        error = f'Error in {func_name}: queue is empty()...'
                        logger.error(error)
                        database.update_worker_task({'status': 'error', 'result': error}, self._current_task,
                                                    self._worker)

        except Exception as e:
            exception = e
            error = f'Exception in {func_name}: {type(e)}: {str(e)}; traceback: {traceback.format_exc()}'
        finally:
            if timer:
                timer.cancel()
            if exception:
                if self._raise_exception:
                    if self._worker:
                        database.set_worker_error(error, self._worker, self._current_task)
                    raise exception
                else:
                    database.update_worker_task({'status': 'error', 'result': error}, self._current_task,
                                                self._worker)
                    logger.error(error)
            if not self._blocking:
                if self._life_timeout_sec is not None:
                    time_delta = datetime.datetime.now() - self._started_timestamp
                    time_delta_sec = time_delta.total_seconds()
                    if time_delta_sec > self._life_timeout_sec:
                        logger.info(f'{func_name}; time_delta_sec({time_delta_sec})'
                                    f' > life_timeout_sec({self._life_timeout_sec}) => return...')
                        if self._worker:
                            self._worker.status = 'finished'
                            self._worker.finished_time = datetime.datetime.now()
                            database.session.commit()
                        return
                threading.Timer(self._sleep_sec, self._do_task, args=(process_type,)).start()

    def _do_current_task(self, process_type):
        func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
        # print(f'{func_name}; os.getpid(): {os.getpid()}; os.getppid(): {os.getppid()}')
        if not self._started:
            logger.warning(f'{func_name}; self._started = False; return')
            return

        exception = None
        error = None
        timer = None
        try:
            # print(f'{func_name}: now={datetime.datetime.now()}; task: {self._current_task}')
            task_dict = database.get_record_dict(self._current_task)
            if process_type is None:
                # timer = threading.Timer(self._timeout_sec, self.stop_woker, args=(func_name,))
                # timer.setDaemon(False)
                # timer.start()
                result_dict = self.worker(task_dict)
                # timer.cancel()
                database.update_worker_task(result_dict, self._current_task, self._worker)
            elif process_type in ['fork', 'thread', 'kthread']:
                if process_type == 'fork':
                    worker_queue = multiprocessing.Queue()
                    worker = multiprocessing.Process(target=self.worker_process, args=(worker_queue, task_dict,))
                else:
                    worker_queue = queue.Queue()
                    if process_type == 'thread':
                        worker = threading.Thread(target=self.worker_process, args=(worker_queue, task_dict,))
                        worker.setDaemon(True)
                    else:
                        worker = kthread(target=self.worker_process, args=(worker_queue, task_dict,))
                worker.start()
                worker.join(timeout=self._timeout_sec)
                if worker.is_alive():
                    if process_type in ['fork', 'kthread']:
                        worker.terminate()
                        error = f'Error in {func_name}: worker({process_type}).is_alive() => worker.terminate()'
                        logger.error(error)
                        database.update_worker_task({'status': 'error', 'result': error}, self._current_task,
                                                    self._worker)
                    else:
                        error = f'Error in {func_name}: worker({process_type}).is_alive() => sys.exit(1)'
                        logger.error(error)
                        if self._worker:
                            database.set_worker_error(error, self._worker, self._current_task)
                        sys.exit(1)
                else:
                    if not worker_queue.empty():
                        database.update_worker_task(worker_queue.get_nowait(), self._current_task, self._worker)
                    else:
                        error = f'Error in {func_name}: queue is empty()...'
                        logger.error(error)
                        database.update_worker_task({'status': 'error', 'result': error}, self._current_task,
                                                    self._worker)

        except Exception as e:
            exception = e
            error = f'Exception in {func_name}: {type(e)}: {str(e)}; traceback: {traceback.format_exc()}'
        finally:
            if timer:
                timer.cancel()
            if exception:
                if self._raise_exception:
                    if self._worker:
                        database.set_worker_error(error, self._worker, self._current_task)
                    database.close_session()
                    raise exception
                else:
                    database.update_worker_task({'status': 'error', 'result': error}, self._current_task,
                                                self._worker)
                    logger.error(error)

    def _do_tasks(self, process_type, task_limit):
        func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
        if not self._started:
            logger.warning(f'{func_name}; self._started = False; return')
            return

        if not database.get_session():
            logger.error(f'Error in {func_name}: not database.get_session()...')
            return

        new_tasks = database.get_new_tasks(task_limit, self._worker)
        if isinstance(self._current_task, str):
            logger.error(f'Error in {func_name}; {new_tasks}')
            return

        for task in new_tasks:
            task.started_time = datetime.datetime.now()
            if self._worker:
                self._worker.current_task_id = task.task_id
            database.session.commit()
            self._current_task = task
            self._do_current_task(process_type)

        if self._worker:
            self._worker.current_task_id = None

        database.session.commit()
        database.close_session()

        if not self._blocking:
            if self._life_timeout_sec is not None:
                time_delta = datetime.datetime.now() - self._started_timestamp
                time_delta_sec = time_delta.total_seconds()
                if time_delta_sec > self._life_timeout_sec:
                    logger.info(f'{func_name}; time_delta_sec({time_delta_sec})'
                                f' > life_timeout_sec({self._life_timeout_sec}) => return...')
                    if self._worker:
                        self._worker.status = 'finished'
                        self._worker.finished_time = datetime.datetime.now()
                        database.session.commit()
                    return
            threading.Timer(self._sleep_sec, self._do_task, args=(process_type,)).start()

    def start(self, cfg=None, config_id=None):
        func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'

        if cfg is None:
            if not isinstance(config.cfg, dict):
                logger.error(f'Error in {func_name}: not isinstance(config.cfg, dict)...')
                return
            cfg = config.cfg

        if not database.init(cfg.get('database')):
            logger.error(f'Error in {func_name}: not database.init()...')
            return

        if not database.get_session():
            error = f'Error in {func_name}: not database.get_session()...'
            logger.error(error)
            return error

        process_type = None
        task_limit = 20
        loops_limit = 1
        create_db_worker = False
        worker_cfg = cfg.get('worker')

        if not isinstance(worker_cfg, dict):
            worker_cfg = dict()

        if 'sleep_sec' in worker_cfg:
            worker_cfg['sleep_sec'] = float(worker_cfg.get('sleep_sec'))
        if 'timeout_sec' in worker_cfg:
            worker_cfg['timeout_sec'] = float(worker_cfg.get('timeout_sec'))
        if 'life_timeout_sec' in worker_cfg:
            worker_cfg['life_timeout_sec'] = float(worker_cfg.get('life_timeout_sec'))
        if 'loops_limit' in worker_cfg:
            worker_cfg['loops_limit'] = int(worker_cfg.get('loops_limit'))
        if 'task_limit' in worker_cfg:
            worker_cfg['task_limit'] = int(worker_cfg.get('task_limit'))

        if config_id is not None:
            db_cfg = database.session.query(database.Config).filter(
                database.Config.config_id == config_id).first()
            if db_cfg is None:
                db_cfg = database.Config(config_id=config_id, config=worker_cfg)
                database.add_new_record(db_cfg, database.Config)
            else:
                for key in worker_cfg.keys():
                    if key in db_cfg.config:
                        worker_cfg[key] = db_cfg.config.get(key)
                for key in db_cfg.config.keys():
                    if key not in worker_cfg:
                        worker_cfg[key] = db_cfg.config.get(key)

        if 'enabled' in worker_cfg:
            if worker_cfg.get('enabled', True) is False:
                logger.warning(f'Error in {func_name}: enabled if False => return...')
                database.close_session()
                return

        if 'blocking' in worker_cfg:
            self._blocking = worker_cfg.get('blocking', self._blocking)
        if 'raise_exception' in worker_cfg:
            self._raise_exception = worker_cfg.get('raise_exception', self._raise_exception)
        if 'sleep_sec' in worker_cfg:
            self._sleep_sec = worker_cfg.get('sleep_sec')
        if 'timeout_sec' in worker_cfg:
            self._timeout_sec = worker_cfg.get('timeout_sec')
        if 'life_timeout_sec' in worker_cfg:
            self._life_timeout_sec = worker_cfg.get('life_timeout_sec')
        if 'loops_limit' in worker_cfg:
            loops_limit = worker_cfg.get('loops_limit')
        if 'task_limit' in worker_cfg:
            task_limit = worker_cfg.get('task_limit')

        if loops_limit is not None:
            self._blocking = True
        if 'process_type' in worker_cfg:
            process_type = worker_cfg.get('process_type')

        if process_type is not None:
            if process_type.lower() not in ['fork', 'thread', 'kthread']:
                logger.error(f'Error in {func_name}: unknown process_type "{process_type}"...')
                return

        if 'project' in worker_cfg:
            database.project = worker_cfg.get('project')

        if create_db_worker:
            self._worker = database.get_new_worker()

        database.close_session()

        self._started = True
        self._started_timestamp = datetime.datetime.now()
        if self._blocking:
            loops_counter = 0
            while self._started:
                self._do_tasks(process_type, task_limit)
                loops_counter += 1
                if loops_limit and loops_counter >= loops_limit:
                    logger.info(f'{func_name}; loops_counter({loops_counter})'
                                f' == loops_limit({loops_limit}) => break...')
                    break
                if self._life_timeout_sec is not None:
                    time_delta = datetime.datetime.now() - self._started_timestamp
                    time_delta_sec = time_delta.total_seconds()
                    if time_delta_sec > self._life_timeout_sec:
                        logger.info(f'{func_name}; time_delta_sec({time_delta_sec})'
                                    f' > life_timeout_sec({self._life_timeout_sec}) => break...')
                        break
                if self._sleep_sec is not None:
                    time.sleep(self._sleep_sec)
            if self._worker:
                self._worker.status = 'finished'
                self._worker.finished_time = datetime.datetime.now()
                database.session.commit()
        else:
            self._do_tasks(process_type, task_limit)
            # self._do_task(process_type)

    def stop(self):
        self._started = False


if __name__ == '__main__':
    name = 'Worker'
    ver = '1.3'
    release_date = '2020-06-08 06:00'
    version = f'{name}. ver: {ver} ({release_date})'
    from pg_tasks_queue.Logging import init_logger
    init_logger()

    logger.warning(f'Start {version}')
    Worker().start(config.cfg, config_id=1)
    logger.warning(f'End {version}')





import datetime
import socket
import json
from pg_tasks_queue.Config import cfg as config

from sqlalchemy import create_engine
from sqlalchemy import schema as sqlalchemy_schema
from sqlalchemy import (Boolean, Column, DateTime, ForeignKey, Integer, Text, VARCHAR)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship
from sqlalchemy.orm import sessionmaker
# from sqlalchemy_utils import database_exists
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import and_
from sqlalchemy import or_
# from sqlalchemy.ext.declarative import declared_attr
# from sqlalchemy import MetaData


import logging
logger = logging.getLogger(__name__)

# config.set_config_dir(os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', 'conf'))

database_cfg = None
db_engine = None
session = None
db_schema = None
project = None

# database_cfg = config.cfg.get('database')
# if isinstance(database_cfg, dict):
#     db_schema = database_cfg.get('schema')


def get_repr(_class):
    repr = f'<{_class.__class__.__name__}('
    for i in range(len(_class.__table__.columns)):
        column = list(_class.__table__.columns)[i]
        val = getattr(_class, column.name)
        repr += f'{column.name}={val}'
        if i < len(_class.__table__.columns) - 1:
            repr += ', '
    repr += ')>'
    return repr


# # https://docs.sqlalchemy.org/en/13/orm/extensions/declarative/mixins.html
# class Base(object):
#     _schema = None
#     # if db_schema:
#     #     _schema = db_schema
#
#     @declared_attr
#     def __table_args__(cls):
#         args = dict()
#         if cls._schema:
#             args.update({'schema': cls._schema})
#         print(f'CronjobTask: args: {args}')
#         return args
# Base = declarative_base(cls=Base)
# Base._schema = db_schema
# Base = declarative_base(metadata=MetaData(schema=db_schema))

Base = declarative_base()


class CronjobTask(Base):
    __tablename__ = 'cronjob'
    # if db_schema:
    #     __table_args__ = {'schema': db_schema}

    task_id = Column(Integer, primary_key=True)
    module = Column(VARCHAR, nullable=False)
    func = Column(VARCHAR, nullable=False)
    params = Column(JSONB)
    project = Column(VARCHAR)
    max_retry_count = Column(Integer)
    schedule = Column(VARCHAR)
    enabled = Column(Boolean)
    last_start_time = Column(DateTime)
    next_start_time = Column(DateTime)

    def __repr__(self):
        return get_repr(self)


class Worker(Base):
    __tablename__ = 'worker'

    worker_id = Column(Integer, primary_key=True)
    project = Column(VARCHAR)
    host = Column(VARCHAR)
    status = Column(VARCHAR)
    started_time = Column(DateTime)
    finished_time = Column(DateTime)
    success_tasks = Column(Text)
    failed_tasks = Column(Text)
    current_task_id = Column(Integer)
    error = Column(Text)

    def __repr__(self):
        return get_repr(self)


class Task(Base):
    __tablename__ = 'task'

    task_id = Column(Integer, primary_key=True)
    module = Column(VARCHAR, nullable=False)
    func = Column(VARCHAR, nullable=False)
    params = Column(JSONB)
    project = Column(VARCHAR)
    result = Column(Text)
    status = Column(VARCHAR)
    retry_count = Column(Integer)
    max_retry_count = Column(Integer)
    scheduled_time = Column(DateTime)
    defer_time = Column(DateTime)
    create_time = Column(DateTime)
    started_time = Column(DateTime)
    finished_time = Column(DateTime)
    priority = Column(Integer)
    worker_host = Column(VARCHAR)
    failed_email = Column(VARCHAR)
    success_email = Column(VARCHAR)
    parent_task_id = Column(Integer)
    cronjob_task_id = Column(Integer, ForeignKey(CronjobTask.task_id))
    worker_id = Column(Integer, ForeignKey(Worker.worker_id))

    cronjob_task = relationship("CronjobTask")
    worker = relationship("Worker")

    def __repr__(self):
        return get_repr(self)


class Config(Base):

    __tablename__ = 'config'

    config_id = Column(Integer, primary_key=True)
    config = Column(JSONB)

    def __repr__(self):
        return get_repr(self)

# ======================================================================================================================


def get_connection_string():
    host = database_cfg.get('host')
    if host is None:
        logger.error(f"Error in Database.get_connection_string(): host is None...")
        return None
    dbname = database_cfg.get('dbname')
    if dbname is None:
        logger.error(f"Error in Database.get_connection_string(): dbname is None...")
        return None
    user = database_cfg.get('user')
    if user is None:
        logger.error(f"Error in Database.get_connection_string(): user is None...")
        return None
    password = database_cfg.get('password')
    if password is None:
        logger.error(f"Error in Database.get_connection_string(): password is None...")
        return None
    port = database_cfg.get('port')

    schema = database_cfg.get('schema')
    if schema is None:
        logger.error(f"Error in Database.get_connection_string(): schema is None...")
        return None
    if port:
        connection_string = f'postgresql+psycopg2://{user}:{password}@{host}:{port}/{dbname}'
    else:
        connection_string = f'postgresql+psycopg2://{user}:{password}@{host}/{dbname}'
    return connection_string


def init(cfg=None, use_project=True, create=True):

    if cfg is None:
        if not isinstance(config.cfg, dict):
            logger.error(f"Error in Database.init(): not isinstance(config.cfg, dict)...")
            return False
        cfg = config.cfg.get('database')
    global database_cfg
    database_cfg = cfg

    if not isinstance(database_cfg, dict):
        logger.error(f'Error in Database.init(): not isinstance(database_cfg, dict)...')
        return False

    if use_project:
        global project
        project = database_cfg.get('project')
        if not isinstance(project, str):
            logger.error(f'Error in Database.init(): not isinstance(project, str)...')
            return False

    global db_schema
    db_schema = database_cfg.get('schema')
    if not isinstance(db_schema, str):
        logger.error(f'Error in Database.init(): not isinstance(db_schema, str)...')
        return False

    connection_string = get_connection_string()
    if connection_string is None:
        logger.error(f"Error in Database.init(): connection_string is None...")
        return False

    # if not database_exists(connection_string):
    #     logger.error(f"Error in Database.init(): database '{database_cfg.get('dbname')}' not exists ...")
    #     return False

    global db_engine
    # connect_args = {'options': f'-csearch_path={db_schema}'}
    # db_engine = create_engine(connection_string, echo=database_cfg.get('echo', False), connect_args=connect_args)
    db_engine = create_engine(connection_string, echo=database_cfg.get('echo', False))

    if not db_engine.dialect.has_schema(db_engine, db_schema):
        if not create:
            logger.error(f"Error in Database.init(): schema '{db_schema}' not exists in database "
                         f"'{database_cfg.get('dbname')}'...")
            return False
        logger.warning(f"Try to create schema '{db_schema}' for database {database_cfg.get('dbname')}'...")
        db_engine.execute(sqlalchemy_schema.CreateSchema(db_schema))
        if not db_engine.dialect.has_schema(db_engine, db_schema):
            logger.error(f"Error in Database.init(): schema '{db_schema}' not exists in database "
                         f"'{database_cfg.get('dbname')}'...")
            return False
        logger.warning(f"Schema '{db_schema}' created successfully...")

    global Base
    # Base.metadata.schema = db_schema
    for table_name in Base.metadata.tables:
        Base.metadata.tables[table_name].schema = db_schema

    if create:
        Base.metadata.create_all(db_engine)

    return True


def get_table_columns(_table):
    columns = dict()
    for column in _table.__table__.columns:
        columns[column.name] = column.type.python_type.__name__
    return columns


def update_elem(elem, **kwargs):
    for key, value in kwargs.items():
        if hasattr(elem, key):
            setattr(elem, key, value)


def get_record_dict(_record):
    _dict = dict()
    for column in _record.__table__.columns:
        _dict[column.name] = getattr(_record, column.name)
    return _dict


def get_html_table(query_list, columns=None, table=None):
    if len(query_list) == 0:
        if table is None:
            return None
        types = get_table_columns(table)
    else:
        types = get_table_columns(query_list[0])
    if columns is None:
        columns = list(types.keys())
    else:
        keys = list(types.keys())
        for key in keys:
            if key not in columns:
               del types[key]

    values = list()
    for row in query_list:
        row_dict = dict()
        for column in columns:
            value = getattr(row, column)
            value_type = types.get(column)
            if value_type == 'datetime' and value is not None:
                value = value.strftime('%Y-%m-%d %H:%M:%S')
            row_dict[column] = value
        values.append(row_dict)
    return {'columns': columns, 'values': values, 'types': types}


def get_session(autocommit=False):
    global session
    if session is None:
        if db_engine is None:
            logger.error(f'Error in Database.get_session(): db_engine is None...')
            return False
        try:
            Session = sessionmaker(bind=db_engine)
            session = Session()
        except Exception as e:
            logger.exception(f'Exception in Database.get_session(): {type(e)}: {str(e)}...')
            session = None
    if session is None:
        logger.error(f'Error in Database.get_session(): session is None...')
        return False
    session.autocommit = autocommit
    return True


def close_session():
    global session
    try:
        if session is not None:
            session.close()
    except Exception as e:
        logger.exception(f'Exception in Database.close_session(): {type(e)}: {str(e)}...')
    finally:
        session = None


def get_table(_value, _class):
    if isinstance(_value, dict):
        _value = _class(**_value)
    return _value


def add_common_columns(_value):
    if 'project' in _value.__table__.columns:
        if _value.project is None:
            _value.project = project
    if 'params' in _value.__table__.columns:
        if _value.params is None:
            _value.params = dict()
    if 'create_time' in _value.__table__.columns:
        if _value.create_time is None:
            _value.create_time = datetime.datetime.now()
    return _value


def add_new_record(_add_value, _class, func=None):
    if session is None:
        error = f'Error in Database.add_new_record(): session is None...'
        logger.error(error)
        return error
    _add_value = get_table(_add_value, _class)
    if not isinstance(_add_value, _class):
        error = f'Error in Database.add_new_record(): not isinstance(_add_value, {_class.__class__.__name__})...'
        logger.error(error)
        return error
    if func is not None:
        _add_value.module = func.__module__
        _add_value.func = func.__name__

    session.add(add_common_columns(_add_value))
    session.commit()
    return _add_value


def add_cornjob_task(_add_value, func=None, commit=True):
    return add_new_record(_add_value, CronjobTask, func=func)


def add_task(_add_value, func=None):
    return add_new_record(_add_value, Task, func=func)


def get_new_worker():
    host_name = socket.gethostname()
    host_ip = socket.gethostbyname(host_name)
    host = f'{host_ip} : {host_name}'
    started_time = datetime.datetime.now()
    success_tasks = json.dumps(list())
    failed_tasks = json.dumps(list())
    new_worker = Worker(host=host, status='working', started_time=started_time,
                        success_tasks=success_tasks, failed_tasks=failed_tasks)
    return add_new_record(new_worker, Worker)


def get_new_task(worker=None):
    if session is None:
        error = f'Error in Database.get_new_task(): session is None...'
        logger.error(error)
        return error

    # 1. status is null and scheduled_time <= NOW()
    filter_value = and_(Task.status.is_(None),
                        Task.scheduled_time.isnot(None),
                        Task.scheduled_time <= datetime.datetime.now())
    if project:
        filter_value = and_(filter_value, Task.project == project)
    task = session.query(Task).with_for_update(skip_locked=True).filter(filter_value).order_by(
        Task.scheduled_time.asc(), Task.priority.desc()).first()

    if task is None:
        # 2. status = 'error' max_retry_count > retry_count
        filter_value = and_(Task.status == 'error',
                            or_(Task.retry_count.is_(None), Task.max_retry_count > Task.retry_count))
        if project:
            filter_value = and_(filter_value, Task.project == project)
        task = session.query(Task).with_for_update(skip_locked=True).filter(filter_value).order_by(
            Task.finished_time.asc(), Task.priority.desc()).first()

    if task is None:
        # 3. Simple Tasks
        filter_value = and_(Task.status.is_(None), Task.scheduled_time.is_(None))
        if project:
            filter_value = and_(filter_value, Task.project == project)
        task = session.query(Task).with_for_update(skip_locked=True).filter(filter_value).order_by(
            Task.priority.desc(), Task.task_id.asc()).first()

    if task is not None:
        if worker is None:
            host_name = socket.gethostname()
            host_ip = socket.gethostbyname(host_name)
            task.worker_host = f'{host_ip} : {host_name}'
        else:
            task.worker_host = worker.host
            task.worker_id = worker.worker_id
            worker.current_task_id = task.task_id
        task.started_time = datetime.datetime.now()
        task.status = 'working'

    session.commit()
    return task


def get_new_tasks(task_limit, worker=None):
    if session is None:
        error = f'Error in Database.get_new_task(): session is None...'
        logger.error(error)
        return error

    # 1. status is null and scheduled_time <= NOW()
    filter_value_1 = and_(Task.status.is_(None),
                          Task.scheduled_time.isnot(None),
                          Task.scheduled_time <= datetime.datetime.now())

    # 2. status = 'error' max_retry_count > retry_count
    filter_value_2 = and_(Task.status == 'error',
                          or_(Task.retry_count.is_(None), Task.max_retry_count > Task.retry_count))

    # 3. Simple Tasks
    filter_value_3 = and_(Task.status.is_(None), Task.scheduled_time.is_(None))

    filter_value = or_(filter_value_1, filter_value_2, filter_value_3)

    if project:
        filter_value = and_(filter_value, Task.project == project)

    query = session.query(Task).with_for_update(skip_locked=True).filter(filter_value).order_by(
        Task.task_id.asc()).limit(task_limit)

    tasks = query.all()
    for task in tasks:
        if worker is None:
            host_name = socket.gethostname()
            host_ip = socket.gethostbyname(host_name)
            task.worker_host = f'{host_ip} : {host_name}'
        else:
            task.worker_host = worker.host
            task.worker_id = worker.worker_id
            # worker.current_task_id = task.task_id
        # task.started_time = datetime.datetime.now()
        task.status = 'working'

    session.commit()
    return tasks


def update_worker_task(_result, task, worker):
    if session is None:
        error = f'Error in Database.update_worker_task(): session is None...'
        logger.error(error)
        return error

    finished_time = datetime.datetime.now()
    error = None
    if isinstance(_result, dict):
        status = _result.get('status')
        result = _result.get('result', '')
        if not isinstance(status, str):
            error = f'Error in Database.update_worker_task(): not isinstance(status, strt)...'
            logger.error(error)
        elif status == 'error':
            error = result
    else:
        status = 'finished'
        result = _result

    if error is not None:
        task.status = 'error'
        task.result = error

        max_retry_count = task.max_retry_count
        max_retry_count = 0 if max_retry_count is None else max_retry_count
        if max_retry_count > 0:
            retry_count = task.retry_count
            retry_count = 0 if retry_count is None else retry_count
            task.retry_count = retry_count + 1

        if worker:
            failed_tasks = json.loads(worker.failed_tasks)
            failed_tasks.append(task.task_id)
            worker.failed_tasks = json.dumps(failed_tasks)

    else:
        task.status = status
        task.result = result

        if worker:
            success_tasks = json.loads(worker.success_tasks)
            success_tasks.append(task.task_id)
            worker.success_tasks = json.dumps(success_tasks)

    if worker:
        worker.current_task_id = None

    task.finished_time = finished_time
    session.commit()

    return None


def set_worker_error(error, worker, task):
    if session is None:
        error = f'Error in Database.set_worker_error(): session is None...'
        logger.error(error)
        return error

    finished_time = datetime.datetime.now()

    if task:
        task.status = 'error'
        task.result = error
        task.finished_time = finished_time

        max_retry_count = task.max_retry_count
        max_retry_count = 0 if max_retry_count is None else max_retry_count
        if max_retry_count > 0:
            retry_count = task.retry_count
            retry_count = 0 if retry_count is None else retry_count
            task.retry_count = retry_count + 1

        failed_tasks = json.loads(worker.failed_tasks)
        failed_tasks.append(task.task_id)
        worker.failed_tasks = json.dumps(failed_tasks)

    worker.status = 'error'
    worker.error = error
    worker.finished_time = finished_time

    session.commit()

# ======================================================================================================================


if __name__ == '__main__':
    name = 'Database'
    ver = '1.2'
    release_date = '2020-06-05 10:00'
    version = f'{name}. ver: {ver} ({release_date})'
    from pg_tasks_queue.Logging import init_logger
    init_logger()
    logger.warning(version)

    cfg = config.cfg.get('database')

    if not init(cfg):
        logger.warning(f"not init()")
    else:
        if not get_session():
            logger.warning(f"not get_session()")
        else:
            # get_new_task()

            new_cornjob_task = {'schedule': '*/1 * * * *',
                                'module': 'Modules.TestTasks',
                                'func': 'raise_exception',  # raise_exception, sleep_false, sleep
                                'params': {"counter": 4},
                                'project': 'project_2',
                                'max_retry_count': 2,
                                'enabled': True}
            # new_cornjob_task = CronjobTask(**new_cornjob_task)
            # from Modules import TestTasks
            # new_cornjob_task = add_cornjob_task(new_cornjob_task, TestTasks.raise_exception)
            new_cornjob_task = add_cornjob_task(new_cornjob_task)
            print(f'new_cornjob_task: {new_cornjob_task}')

            import datetime
            new_task = {'module': 'Modules.TestTasks',
                        'func': 'raise_exception',
                        'params': {"counter": 4},
                        'max_retry_count': 20,
                        'scheduled_time': datetime.datetime.now(),
                        'priority': 1}
            new_task = add_task(new_task)
            print(f'new_task: {new_task}')

            close_session()



import logging
import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration
from pg_tasks_queue.Config import cfg as config


def get_log_level(level, def_value=None):
    if not isinstance(level, str):
        return def_value
    level = level.upper()
    return getattr(logging, level) if isinstance(getattr(logging, level), int) else def_value


log_level = logging.INFO
log_format = '%(asctime)s-%(levelname)s-%(module)s:%(lineno)d - %(message)s'
log_filemode = 'a'

logging_cfg = config.cfg.get('logging')
if isinstance(logging_cfg, dict):
    log_filemode = logging_cfg.get('filemode', log_filemode)
    if isinstance(logging_cfg.get('format'), str):
        log_format = logging_cfg.get('format').replace('&', '%')

    log_level = get_log_level(logging_cfg.get('level'), log_level)
    sentry_sdk_dsn = logging_cfg.get('sentry_sdk_dsn')
    if isinstance(sentry_sdk_dsn, str):
        if sentry_sdk_dsn.lower() != 'none':
            sentry_logging = LoggingIntegration(
                level=logging.INFO,  # Capture info and above as breadcrumbs
                event_level=logging.ERROR  # Send errors as events
            )
            sentry_sdk.init(dsn=sentry_sdk_dsn, integrations=[sentry_logging])


def init_logger(log_filename=None):
    if log_filename is None:
        log_filename = logging_cfg.get('file')
    if isinstance(log_filename, str):
        log_filename = None if (log_filename.lower() == 'none' or log_filename.strip() == '') else log_filename
    logging.basicConfig(format=log_format, level=log_level)
    if log_filename is not None:
        file_handler = logging.FileHandler(log_filename, mode=log_filemode)
        file_handler.setFormatter(logging.Formatter(log_format))
        file_handler.setLevel(log_level)
        logging.getLogger().addHandler(file_handler)




import sys
from croniter import croniter
import datetime
import time
from sqlalchemy import and_
from pg_tasks_queue.Config import cfg as config
from pg_tasks_queue.Logging import init_logger
import pg_tasks_queue.Database as database

import logging
logger = logging.getLogger(__name__)


class CronJobTasks:

    _started = False

    def update_tasks(self):
        func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'

        now = datetime.datetime.now()
        logger.info(f'Start {func_name} at {now}')

        if not database.get_session():
            error = f'Error in {func_name}: not database.get_session()...'
            logger.error(error)
            return error

        if database.project:
            filter_value = and_(database.CronjobTask.project == database.project,
                                database.CronjobTask.enabled.is_(True))
        else:
            filter_value = database.CronjobTask.enabled.is_(True)

        cronjob_tasks = database.session.query(database.CronjobTask).filter(filter_value).all()
        database.session.commit()

        if not isinstance(cronjob_tasks, list):
            error = f'Error in {func_name}: not isinstance(cronjob_tasks, list)...'
            logger.error(error)
            return error

        start_of_min = datetime.datetime.strptime(now.strftime('%Y-%m-%d %H:%M:00'), '%Y-%m-%d %H:%M:00')

        for cronjob_task in cronjob_tasks:

            itr = croniter(cronjob_task.schedule, start_of_min)
            scheduled_time = itr.get_next(datetime.datetime)

            if database.project:
                filter_value = and_(database.Task.project == database.project,
                                    database.Task.cronjob_task_id == cronjob_task.task_id,
                                    database.Task.scheduled_time > start_of_min)

            else:
                filter_value = and_(database.Task.cronjob_task_id == cronjob_task.task_id,
                                    database.Task.scheduled_time > start_of_min)

            task = database.session.query(database.Task).filter(filter_value).first()
            database.session.commit()

            if task is None:
                logger.info(f'{func_name}; apply cronjob_task   : {cronjob_task}')
                cronjob_task.last_start_time = cronjob_task.next_start_time
                cronjob_task.next_start_time = scheduled_time
                database.session.commit()
                logger.info(f'{func_name}; updated cronjob_task : {cronjob_task}')
                new_task_dict = {'priority': 1,
                                 'create_time': start_of_min,
                                 'scheduled_time': scheduled_time,
                                 'cronjob_task_id': cronjob_task.task_id}
                task_columns = list(database.get_table_columns(database.Task()).keys())
                task_columns.remove('task_id')
                for task_column in task_columns:
                    if task_column in cronjob_task.__table__.columns:
                        new_task_dict[task_column] = getattr(cronjob_task, task_column)
                new_task = database.add_task(new_task_dict)
                logger.info(f'{func_name}; create new task      : {new_task}')

        database.close_session()

        return None

    def start(self, cfg=None):
        func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'

        raise_exception = False
        try:

            if cfg is None:
                if not isinstance(config.cfg, dict):
                    logger.error(f'Error in {func_name}: not isinstance(config.cfg, dict)...')
                    return
                cfg = config.cfg

            if not database.init(cfg.get('database')):
                logger.error(f'Error in {func_name}: not database.init()...')
                return

            sleep_sec = None
            loops_limit = 1
            life_timeout_sec = None

            cronjobs_cfg = cfg.get('cronjobs')
            if isinstance(cronjobs_cfg, dict):
                raise_exception = cronjobs_cfg.get('raise_exception', raise_exception)

                if cronjobs_cfg.get('sleep_sec') is not None:
                    sleep_sec = float(cronjobs_cfg.get('sleep_sec'))

                if cronjobs_cfg.get('life_timeout_sec') is not None:
                    life_timeout_sec = float(cronjobs_cfg.get('life_timeout_sec'))

                if cronjobs_cfg.get('loops_limit') is not None:
                    loops_limit = int(cronjobs_cfg.get('loops_limit'))

            started_timestamp = datetime.datetime.now()
            loops_counter = 0
            self._started = True
            while self._started:
                self.update_tasks()
                loops_counter += 1
                if loops_limit and loops_counter >= loops_limit:
                    logger.info(f'{func_name}; loops_counter({loops_counter})'
                                f' == loops_limit({loops_limit}) => break...')
                    break
                elif life_timeout_sec is not None:
                    time_delta = datetime.datetime.now() - started_timestamp
                    time_delta_sec = time_delta.total_seconds()
                    if time_delta_sec > life_timeout_sec:
                        logger.info(f'{func_name}; time_delta_sec({time_delta_sec})'
                                    f' > life_timeout_sec({life_timeout_sec}) => break...')
                        break
                time.sleep(sleep_sec)

        except Exception as e:
            logger.exception(f'Error in {func_name}: {type(e)}: {str(e)}...')
            if raise_exception:
                raise e

    def stop(self):
        self._started = False


if __name__ == '__main__':
    name = 'CronJob Tasks'
    ver = '1.2'
    release_date = '2020-06-05 10:25'
    version = f'{name}. ver: {ver} ({release_date})'
    init_logger()

    logger.warning(f'Start {version}')
    CronJobTasks().start(config.cfg)
    logger.warning(f'End {version}')



import configparser
import os
import sys
import copy
import socket


class Config:

    cfg = None
    conf_dir = None

    def __init__(self):
        conf_dir = os.path.join(os.path.abspath(os.curdir), 'conf')
        if not (os.path.exists(conf_dir) and os.path.isdir(conf_dir)):
            conf_dir = os.path.join(os.path.abspath(os.curdir), '..', 'conf')
        # if not (os.path.exists(conf_dir) and os.path.isdir(conf_dir)):
        #     conf_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', 'conf')
        self.set_config_dir(conf_dir)

    def set_config_dir(self, conf_dir, force=False):
        if os.path.exists(conf_dir) and os.path.isdir(conf_dir):
            if self.cfg is None or force:
                config = None
                conf_base_dir = os.path.join(conf_dir, 'base')
                for name in os.listdir(conf_base_dir):
                    if name.startswith('.'):
                        continue
                    conf_file = os.path.join(conf_base_dir, name)
                    config = self._update_config(config, self._read_config(conf_file))
                host_conf_file = os.path.join(os.path.join(conf_dir, 'host'), socket.gethostname() + '.cfg')
                if os.path.exists(host_conf_file) and os.path.isfile(host_conf_file):
                    config = self._update_config(config, self._read_config(host_conf_file))
                self.cfg = config
                self.update_from_env()
                self.update_values()

    def __getitem__(self, item):
        return self.cfg.__getitem__(item)

    def print_confing(self, func_name=None):
        if func_name is None:
            func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
        print(f'{func_name}: config: {self.cfg}')

    @staticmethod
    def _read_config(path):
        config = configparser.ConfigParser()
        config.optionxform = str
        config.read(path)
        return {section: dict(config.items(section)) for section in config.sections()}

    @staticmethod
    def _update_config(config, ext_config):
        if config:
            for key in ext_config:
                if key in config:
                    config[key].update(ext_config[key])
                else:
                    config[key] = ext_config[key]
            return config
        else:
            return ext_config

    def update_from_env(self):
        new_cfg = copy.deepcopy(self.cfg)
        for section_name in self.cfg:
            for key in self.cfg[section_name]:
                env_key = f'{section_name.upper()}_{key.upper()}'
                if env_key in os.environ:
                    new_cfg[section_name][key] = os.environ[env_key]
        self.cfg = new_cfg

    def update_values(self):
        new_cfg = copy.deepcopy(self.cfg)
        for section_name in self.cfg:
            for k, v in self.cfg[section_name].items():
                if isinstance(v, str):
                    if v.strip() == '' or v.strip().lower() == 'none' or v.strip().lower() == '<none>':
                        new_cfg[section_name][k] = None
                    elif v.strip().lower() == 'true':
                        new_cfg[section_name][k] = True
                    elif v.strip().lower() == 'false':
                        new_cfg[section_name][k] = False
        self.cfg = new_cfg


cfg = Config()



from setuptools import setup, find_packages


PACKAGE = "pg_tasks_queue"
NAME = "pg-task-queue"
DESCRIPTION = "Python asinc task system with postgres database"
AUTHOR = "Alex V. Smith"
URL = "https://gitlab.com/cryptoner_dev/tasks_system"
AUTHOR_EMAIL = "smith.it.interface@gmail.com"
VERSION = __import__(PACKAGE).__version__

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    author=AUTHOR,
    url=URL,
    author_email=AUTHOR_EMAIL,
    license="BSD",
    packages=find_packages(exclude=['Admin*', 'Modules']),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    install_requires=[
        'psycopg2',
        'pandas',
        'sqlalchemy',
        'sqlalchemy_utils',
        'croniter',
        'sentry_sdk'
    ],
    include_package_data=True,
    zip_safe=False,
)